# hct/hct_solver.py

import math
from typing import List
from PyMCUlib.utils import color_utils
from PyMCUlib.utils import math_utils
from PyMCUlib.hct.cam16 import Cam16
from PyMCUlib.hct.viewing_conditions import ViewingConditions

class HctSolver:
    """
    A class that solves the HCT equation.
    """
    
    SCALED_DISCOUNT_FROM_LINRGB = [
        [
            0.001200833568784504,
            0.002389694492170889,
            0.0002795742885861124,
        ],
        [
            0.0005891086651375999,
            0.0029785502573438758,
            0.0003270666104008398,
        ],
        [
            0.00010146692491640572,
            0.0005364214359186694,
            0.0032979401770712076,
        ],
    ]

    LINRGB_FROM_SCALED_DISCOUNT = [
        [
            1373.2198709594231,
            -1100.4251190754821,
            -7.278681089101213,
        ],
        [
            -271.815969077903,
            559.6580465940733,
            -32.46047482791194,
        ],
        [
            1.9622899599665666,
            -57.173814538844006,
            308.7233197812385,
        ],
    ]

    Y_FROM_LINRGB = [0.2126, 0.7152, 0.0722]

    CRITICAL_PLANES = [
        0.015176349177441876, 0.045529047532325624, 0.07588174588720938,
        0.10623444424209313,  0.13658714259697685,  0.16693984095186062,
        0.19729253930674434,  0.2276452376616281,   0.2579979360165119,
        0.28835063437139563,  0.3188300904430532,   0.350925934958123,
        0.3848314933096426,   0.42057480301049466,  0.458183274052838,
        0.4976837250274023,   0.5391024159806381,   0.5824650784040898,
        0.6277969426914107,   0.6751227633498623,   0.7244668422128921,
        0.775853049866786,    0.829304845476233,    0.8848452951698498,
        0.942497089126609,    1.0022825574869039,   1.0642236851973577,
        1.1283421258858297,   1.1946592148522128,   1.2631959812511864,
        1.3339731595349034,   1.407011200216447,    1.4823302800086415,
        1.5599503113873272,   1.6398909516233677,   1.7221716113234105,
        1.8068114625156377,   1.8938294463134073,   1.9832442801866852,
        2.075074464868551,    2.1693382909216234,   2.2660538449872063,
        2.36523901573795,     2.4669114995532007,   2.5710888059345764,
        2.6777882626779785,   2.7870270208169257,   2.898822059350997,
        3.0131901897720907,   3.1301480604002863,   3.2497121605402226,
        3.3718988244681087,   3.4967242352587946,   3.624204428461639,
        3.754355295633311,    3.887192587735158,    4.022731918402185,
        4.160988767090289,    4.301978482107941,    4.445716283538092,
        4.592217266055746,    4.741496401646282,    4.893568542229298,
        5.048448422192488,    5.20615066083972,     5.3666897647573375,
        5.5300801301023865,   5.696336044816294,    5.865471690767354,
        6.037501145825082,    6.212438385869475,    6.390297286737924,
        6.571091626112461,    6.7548350853498045,   6.941541251256611,
        7.131223617812143,    7.323895587840543,    7.5195704746346665,
        7.7182615035334345,   7.919981813454504,    8.124744458384042,
        8.332562408825165,    8.543448553206703,    8.757415699253682,
        8.974476575321063,    9.194643831691977,    9.417930041841839,
        9.644347703669503,    9.873909240696694,    10.106627003236781,
        10.342513269534024,   10.58158024687427,    10.8238400726681,
        11.069304815507364,   11.317986476196008,   11.569896988756009,
        11.825048221409341,   12.083451977536606,   12.345119996613247,
        12.610063955123938,   12.878295467455942,   13.149826086772048,
        13.42466730586372,    13.702830557985108,   13.984327217668513,
        14.269168601521828,   14.55736596900856,    14.848930523210871,
        15.143873411576273,   15.44220572664832,    15.743938506781891,
        16.04908273684337,    16.35764934889634,    16.66964922287304,
        16.985093187232053,   17.30399201960269,    17.62635644741625,
        17.95219714852476,    18.281524751807332,   18.614349837764564,
        18.95068293910138,    19.290534541298456,   19.633915083172692,
        19.98083495742689,    20.331304511189067,   20.685334046541502,
        21.042933821039977,   21.404114048223256,   21.76888489811322,
        22.137256497705877,   22.50923893145328,    22.884842241736916,
        23.264076429332462,   23.6469514538663,     24.033477234264016,
        24.42366364919083,    24.817520537484558,   25.21505769858089,
        25.61628489293138,    26.021211842414342,   26.429848230738664,
        26.842203703840827,   27.258287870275353,   27.678110301598522,
        28.10168053274597,    28.529008062403893,   28.96010235337422,
        29.39497283293396,    29.83362889318845,    30.276079891419332,
        30.722335150426627,   31.172403958865512,   31.62629557157785,
        32.08401920991837,    32.54558406207592,    33.010999283389665,
        33.4802739966603,     33.953417292456834,   34.430438229418264,
        34.911345834551085,   35.39614910352207,    35.88485700094671,
        36.37747846067349,    36.87402238606382,    37.37449765026789,
        37.87891309649659,    38.38727753828926,    38.89959975977785,
        39.41588851594697,    39.93615253289054,    40.460400508064545,
        40.98864111053629,    41.520882981230194,   42.05713473317016,
        42.597404951718396,   43.141702194811224,   43.6900349931913,
        44.24241185063697,    44.798841244188324,   45.35933162437017,
        45.92389141541209,    46.49252901546552,    47.065252796817916,
        47.64207110610409,    48.22299226451468,    48.808024568002054,
        49.3971762874833,     49.9904556690408,     50.587870934119984,
        51.189430279724725,   51.79514187861014,    52.40501387947288,
        53.0190544071392,     53.637271562750364,   54.259673423945976,
        54.88626804504493,    55.517063457223934,   56.15206766869424,
        56.79128866487574,    57.43473440856916,    58.08241284012621,
        58.734331877617365,   59.39049941699807,    60.05092333227251,
        60.715611475655585,   61.38457167773311,    62.057811747619894,
        62.7353394731159,     63.417162620860914,   64.10328893648692,
        64.79372614476921,    65.48848194977529,    66.18756403501224,
        66.89098006357258,    67.59873767827808,    68.31084450182222,
        69.02730813691093,    69.74813616640164,    70.47333615344107,
        71.20291564160104,    71.93688215501312,    72.67524319850172,
        73.41800625771542,    74.16517879925733,    74.9167682708136,
        75.67278210128072,    76.43322770089146,    77.1981124613393,
        77.96744375590167,    78.74122893956174,    79.51947534912904,
        80.30219030335869,    81.08938110306934,    81.88105503125999,
        82.67721935322541,    83.4778813166706,     84.28304815182372,
        85.09272707154808,    85.90692527145302,    86.72564993000343,
        87.54890820862819,    88.3767072518277,     89.2090541872801,
        90.04595612594655,    90.88742016217518,    91.73345337380438,
        92.58406282226491,    93.43925555268066,    94.29903859396902,
        95.16341895893969,    96.03240364439274,    96.9059996312159,
        97.78421388448044,    98.6670533535366,     99.55452497210776,
    ]

    @staticmethod
    def _sanitize_radians(angle: float) -> float:
        """
        Sanitizes a small enough angle in radians.

        Args:
            angle: An angle in radians; must not deviate too much from 0.
        
        Returns:
            A coterminal angle between 0 and 2pi.
        """
        return (angle + math.pi * 8) % (math.pi * 2)

    @staticmethod
    def _true_delinearized(rgb_component: float) -> float:
        """
        Delinearizes an RGB component, returning a floating-point number.

        Args:
            rgb_component: 0.0 <= rgb_component <= 100.0, represents linear R/G/B channel
        
        Returns:
            0.0 <= output <= 255.0, color channel converted to regular RGB space
        """
        normalized = rgb_component / 100.0
        delinearized = 0.0
        if normalized <= 0.0031308:
            delinearized = normalized * 12.92
        else:
            delinearized = 1.055 * math.pow(normalized, 1.0 / 2.4) - 0.055
        return delinearized * 255.0

    @staticmethod
    def _chromatic_adaptation(component: float) -> float:
        """
        Performs chromatic adaptation on a color component.

        Args:
            component: The color component to adapt.
        
        Returns:
            The adapted component.
        """
        af = math.pow(abs(component), 0.42)
        return math_utils.signum(component) * 400.0 * af / (af + 27.13)

    @classmethod
    def _hue_of(cls, linrgb: list[float]) -> float:
        """
        Returns the hue of a linear RGB color in CAM16.

        Args:
            linrgb: The linear RGB coordinates of a color.
        
        Returns:
            The hue of the color in CAM16, in radians.
        """
        scaled_discount = math_utils.matrix_multiply(linrgb, cls.SCALED_DISCOUNT_FROM_LINRGB)
        r_a = cls._chromatic_adaptation(scaled_discount[0])
        g_a = cls._chromatic_adaptation(scaled_discount[1])
        b_a = cls._chromatic_adaptation(scaled_discount[2])
        # redness-greenness
        a = (11.0 * r_a + -12.0 * g_a + b_a) / 11.0
        # yellowness-blueness
        b = (r_a + g_a - 2.0 * b_a) / 9.0
        return math.atan2(b, a)

    @classmethod
    def _are_in_cyclic_order(cls, a: float, b: float, c: float) -> bool:
        """
        Checks if three angles are in cyclic order.

        Args:
            a: First angle.
            b: Second angle.
            c: Third angle.
        
        Returns:
            True if the angles are in cyclic order, False otherwise.
        """
        delta_ab = cls._sanitize_radians(b - a)
        delta_ac = cls._sanitize_radians(c - a)
        return delta_ab < delta_ac

    @staticmethod
    def _intercept(source: float, mid: float, target: float) -> float:
        """
        Solves the lerp equation.

        Args:
            source: The starting number.
            mid: The number in the middle.
            target: The ending number.
        
        Returns:
            A number t such that lerp(source, target, t) = mid.
        """
        return (mid - source) / (target - source)

    @staticmethod
    def _lerp_point(source: List[float], t: float, target: List[float]) -> List[float]:
        """
        Performs linear interpolation between two points.

        Args:
            source: The starting point.
            t: The interpolation factor.
            target: The ending point.
        
        Returns:
            The interpolated point.
        """
        return [
            source[0] + (target[0] - source[0]) * t,
            source[1] + (target[1] - source[1]) * t,
            source[2] + (target[2] - source[2]) * t,
        ]

    @classmethod
    def _set_coordinate(cls, source: List[float], coordinate: float,
                        target: List[float], axis: int) -> List[float]:
        """
        Intersects a segment with a plane.

        Args:
            source: The coordinates of point A.
            coordinate: The R-, G-, or B-coordinate of the plane.
            target: The coordinates of point B.
            axis: The axis the plane is perpendicular with. (0: R, 1: G, 2: B)
        
        Returns:
            The intersection point of the segment AB with the plane R=coordinate,
            G=coordinate, or B=coordinate
        """
        t = cls._intercept(source[axis], coordinate, target[axis])
        return cls._lerp_point(source, t, target)

    @staticmethod
    def _is_bounded(x: float) -> bool:
        """
        Checks if a value is between 0 and 100.

        Args:
            x: The value to check.
        
        Returns:
            True if 0 <= x <= 100, False otherwise.
        """
        return 0.0 <= x and x <= 100.0

    @classmethod
    def _nth_vertex(cls, y: float, n: int) -> List[float]:
        """
        Returns the nth possible vertex of the polygonal intersection.

        Args:
            y: The Y value of the plane.
            n: The zero-based index of the point. 0 <= n <= 11.
        
        Returns:
            The nth possible vertex of the polygonal intersection of the y plane and
            the RGB cube, in linear RGB coordinates, if it exists. If this possible
            vertex lies outside of the cube, [-1.0, -1.0, -1.0] is returned.
        """
        k_r = cls.Y_FROM_LINRGB[0]
        k_g = cls.Y_FROM_LINRGB[1]
        k_b = cls.Y_FROM_LINRGB[2]
        coord_a = 0.0 if n % 4 <= 1 else 100.0
        coord_b = 0.0 if n % 2 == 0 else 100.0
        if n < 4:
            g = coord_a
            b = coord_b
            r = (y - g * k_g - b * k_b) / k_r
            if cls._is_bounded(r):
                return [r, g, b]
            else:
                return [-1.0, -1.0, -1.0]
        elif n < 8:
            b = coord_a
            r = coord_b
            g = (y - r * k_r - b * k_b) / k_g
            if cls._is_bounded(g):
                return [r, g, b]
            else:
                return [-1.0, -1.0, -1.0]
        else:
            r = coord_a
            g = coord_b
            b = (y - r * k_r - g * k_g) / k_b
            if cls._is_bounded(b):
                return [r, g, b]
            else:
                return [-1.0, -1.0, -1.0]

    @classmethod
    def _bisect_to_segment(cls, y: float, target_hue: float) -> List[List[float]]:
        """
        Finds the segment containing the desired color.

        Args:
            y: The Y value of the color.
            target_hue: The hue of the color.
        
        Returns:
            A list of two sets of linear RGB coordinates, each corresponding to an
            endpoint of the segment containing the desired color.
        """
        left = [-1.0, -1.0, -1.0]
        right = left
        left_hue = 0.0
        right_hue = 0.0
        initialized = False
        uncut = True
        for n in range(12):
            mid = cls._nth_vertex(y, n)
            if mid[0] < 0:
                continue
            mid_hue = cls._hue_of(mid)
            if not initialized:
                left = mid
                right = mid
                left_hue = mid_hue
                right_hue = mid_hue
                initialized = True
                continue
            if uncut or cls._are_in_cyclic_order(left_hue, mid_hue, right_hue):
                uncut = False
                if cls._are_in_cyclic_order(left_hue, target_hue, mid_hue):
                    right = mid
                    right_hue = mid_hue
                else:
                    left = mid
                    left_hue = mid_hue
        return [left, right]

    @staticmethod
    def _midpoint(a: List[float], b: List[float]) -> List[float]:
        """
        Calculates the midpoint of two points.

        Args:
            a: First point.
            b: Second point.
        
        Returns:
            The midpoint of the two points.
        """
        return [
            (a[0] + b[0]) / 2,
            (a[1] + b[1]) / 2,
            (a[2] + b[2]) / 2,
        ]

    @staticmethod
    def _critical_plane_below(x: float) -> int:
        """
        Finds the critical plane below a given value.

        Args:
            x: The value.
        
        Returns:
            The index of the critical plane below the value.
        """
        return math.floor(x - 0.5)

    @staticmethod
    def _critical_plane_above(x: float) -> int:
        """
        Finds the critical plane above a given value.

        Args:
            x: The value.
        
        Returns:
            The index of the critical plane above the value.
        """
        return math.ceil(x - 0.5)

    @classmethod
    def _bisect_to_limit(cls, y: float, target_hue: float) -> List[float]:
        """
        Finds a color with the given Y and hue on the boundary of the cube.

        Args:
            y: The Y value of the color.
            target_hue: The hue of the color.
        
        Returns:
            The desired color, in linear RGB coordinates.
        """
        segment = cls._bisect_to_segment(y, target_hue)
        left = segment[0]
        left_hue = cls._hue_of(left)
        right = segment[1]
        for axis in range(3):
            if left[axis] != right[axis]:
                l_plane = -1
                r_plane = 255
                if left[axis] < right[axis]:
                    l_plane = cls._critical_plane_below(
                        cls._true_delinearized(left[axis]))
                    r_plane = cls._critical_plane_above(
                        cls._true_delinearized(right[axis]))
                else:
                    l_plane = cls._critical_plane_above(
                        cls._true_delinearized(left[axis]))
                    r_plane = cls._critical_plane_below(
                        cls._true_delinearized(right[axis]))
                for i in range(8):
                    if abs(r_plane - l_plane) <= 1:
                        break
                    else:
                        m_plane = math.floor((l_plane + r_plane) / 2.0)
                        mid_plane_coordinate = cls.CRITICAL_PLANES[m_plane]
                        mid = cls._set_coordinate(left, mid_plane_coordinate, right, axis)
                        mid_hue = cls._hue_of(mid)
                        if cls._are_in_cyclic_order(left_hue, target_hue, mid_hue):
                            right = mid
                            r_plane = m_plane
                        else:
                            left = mid
                            left_hue = mid_hue
                            l_plane = m_plane
        return cls._midpoint(left, right)

    @staticmethod
    def _inverse_chromatic_adaptation(adapted: float) -> float:
        """
        Inversely adapts a color component.

        Args:
            adapted: The adapted component.
        
        Returns:
            The unadapted component.
        """
        adapted_abs = abs(adapted)
        base = max(0, 27.13 * adapted_abs / (400.0 - adapted_abs))
        return math_utils.signum(adapted) * math.pow(base, 1.0 / 0.42)

    @classmethod
    def _find_result_by_j(cls, hue_radians: float, chroma: float, y: float) -> int:
        """
        Finds a color with the given hue, chroma, and Y.

        Args:
            hue_radians: The desired hue in radians.
            chroma: The desired chroma.
            y: The desired Y.
        
        Returns:
            The desired color as a hexadecimal integer, if found; 0 otherwise.
        """
        # Initial estimate of j.
        j = math.sqrt(y) * 11.0
        # ===========================================================
        # Operations inlined from Cam16 to avoid repeated calculation
        # ===========================================================
        viewing_conditions = ViewingConditions.DEFAULT
        t_inner_coeff = 1 / math.pow(1.64 - math.pow(0.29, viewing_conditions.n), 0.73)
        e_hue = 0.25 * (math.cos(hue_radians + 2.0) + 3.8)
        p1 = e_hue * (50000.0 / 13.0) * viewing_conditions.nc * viewing_conditions.ncb
        h_sin = math.sin(hue_radians)
        h_cos = math.cos(hue_radians)
        for iteration_round in range(5):
            # ===========================================================
            # Operations inlined from Cam16 to avoid repeated calculation
            # ===========================================================
            j_normalized = j / 100.0
            alpha = 0.0 if chroma == 0.0 or j == 0.0 else chroma / math.sqrt(j_normalized)
            t = math.pow(alpha * t_inner_coeff, 1.0 / 0.9)
            ac = viewing_conditions.aw * math.pow(
                j_normalized,
                1.0 / viewing_conditions.c / viewing_conditions.z,
            )
            p2 = ac / viewing_conditions.nbb
            gamma = 23.0 * (p2 + 0.305) * t / (23.0 * p1 + 11 * t * h_cos + 108.0 * t * h_sin)
            a = gamma * h_cos
            b = gamma * h_sin
            r_a = (460.0 * p2 + 451.0 * a + 288.0 * b) / 1403.0
            g_a = (460.0 * p2 - 891.0 * a - 261.0 * b) / 1403.0
            b_a = (460.0 * p2 - 220.0 * a - 6300.0 * b) / 1403.0
            r_c_scaled = cls._inverse_chromatic_adaptation(r_a)
            g_c_scaled = cls._inverse_chromatic_adaptation(g_a)
            b_c_scaled = cls._inverse_chromatic_adaptation(b_a)
            linrgb = math_utils.matrix_multiply(
                [r_c_scaled, g_c_scaled, b_c_scaled],
                cls.LINRGB_FROM_SCALED_DISCOUNT,
            )
            # ===========================================================
            # Operations inlined from Cam16 to avoid repeated calculation
            # ===========================================================
            if linrgb[0] < 0 or linrgb[1] < 0 or linrgb[2] < 0:
                return 0
            k_r = cls.Y_FROM_LINRGB[0]
            k_g = cls.Y_FROM_LINRGB[1]
            k_b = cls.Y_FROM_LINRGB[2]
            fnj = k_r * linrgb[0] + k_g * linrgb[1] + k_b * linrgb[2]
            if fnj <= 0:
                return 0
            if iteration_round == 4 or abs(fnj - y) < 0.002:
                if linrgb[0] > 100.01 or linrgb[1] > 100.01 or linrgb[2] > 100.01:
                    return 0
                return color_utils.argb_from_linrgb(linrgb)
            # Iterates with Newton method,
            # Using 2 * fn(j) / j as the approximation of fn'(j)
            j = j - (fnj - y) * j / (2 * fnj)
        return 0

    @classmethod
    def solve_to_int(cls, hue_degrees: float, chroma: float, lstar: float) -> int:
        """
        Finds an sRGB color with the given hue, chroma, and L*, if possible.

        Args:
            hue_degrees: The desired hue, in degrees.
            chroma: The desired chroma.
            lstar: The desired L*.
        
        Returns:
            A hexadecimal representing the sRGB color. The color has sufficiently
            close hue, chroma, and L* to the desired values, if possible; otherwise,
            the hue and L* will be sufficiently close, and chroma will be maximized.
        """
        if chroma < 0.0001 or lstar < 0.0001 or lstar > 99.9999:
            return color_utils.argb_from_lstar(lstar)
        hue_degrees = math_utils.sanitize_degrees_double(hue_degrees)
        hue_radians = hue_degrees / 180 * math.pi
        y = color_utils.y_from_lstar(lstar)
        exact_answer = cls._find_result_by_j(hue_radians, chroma, y)
        if exact_answer != 0:
            return exact_answer
        linrgb = cls._bisect_to_limit(y, hue_radians)
        return color_utils.argb_from_linrgb(linrgb)

    @classmethod
    def solve_to_cam(cls, hue_degrees: float, chroma: float, lstar: float) -> Cam16:
        """
        Finds an sRGB color with the given hue, chroma, and L*, if possible.

        Args:
            hue_degrees: The desired hue, in degrees.
            chroma: The desired chroma.
            lstar: The desired L*.
        
        Returns:
            An CAM16 object representing the sRGB color. The color has sufficiently
            close hue, chroma, and L* to the desired values, if possible; otherwise,
            the hue and L* will be sufficiently close, and chroma will be maximized.
        """
        return Cam16.from_int(cls.solve_to_int(hue_degrees, chroma, lstar))


from setuptools import setup, find_packages

setup(
    name="PyMCUlib",
    version="1.0.2",
    description="Material Color Utilities Python Library",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    author="Wenkang Li",
    author_email="support@deepblue.cc",
    url="https://github.com/wenkang-deepblue/material-color-utilities-python",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    install_requires=[
        "numpy",
        "pillow",
    ],
    python_requires=">=3.12.0",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries",
        "Topic :: Multimedia :: Graphics",
    ],
)



# utils/string_utils.py

from PyMCUlib.utils import color_utils


def hex_from_argb(argb: int) -> str:
    """
    Converts an ARGB color to its hex string representation.
    
    Args:
        argb: ARGB representation of a color.
    
    Returns:
        Hex string representing color, ex. #ff0000 for red.
    """
    r = color_utils.red_from_argb(argb)
    g = color_utils.green_from_argb(argb)
    b = color_utils.blue_from_argb(argb)
    out_parts = [format(r, '02x'), format(g, '02x'), format(b, '02x')]
    
    return '#' + ''.join(out_parts)


def argb_from_hex(hex_str: str) -> int:
    """
    Converts a hex string to its ARGB representation.
    
    Args:
        hex_str: String representing color as hex code. Accepts strings with or
                without leading #, and string representing the color using 3, 6, or 8
                hex characters.
    
    Returns:
        ARGB representation of color.
    
    Raises:
        ValueError: If the hex string has an unexpected format.
    """
    hex_str = hex_str.replace('#', '')
    is_three = len(hex_str) == 3
    is_six = len(hex_str) == 6
    is_eight = len(hex_str) == 8
    
    if not is_three and not is_six and not is_eight:
        raise ValueError(f'unexpected hex {hex_str}')
    
    r = 0
    g = 0
    b = 0
    
    if is_three:
        r = _parse_int_hex(hex_str[0] * 2)
        g = _parse_int_hex(hex_str[1] * 2)
        b = _parse_int_hex(hex_str[2] * 2)
    elif is_six:
        r = _parse_int_hex(hex_str[0:2])
        g = _parse_int_hex(hex_str[2:4])
        b = _parse_int_hex(hex_str[4:6])
    elif is_eight:
        r = _parse_int_hex(hex_str[2:4])
        g = _parse_int_hex(hex_str[4:6])
        b = _parse_int_hex(hex_str[6:8])
    
    return (255 << 24 | (r & 0xff) << 16 | (g & 0xff) << 8 | (b & 0xff)) & 0xFFFFFFFF


def _parse_int_hex(value: str) -> int:
    """
    Parse a hexadecimal string to an integer.
    
    Args:
        value: The hexadecimal string
        
    Returns:
        The integer value
    """
    return int(value, 16)


# scheme/scheme_neutral.py

"""
A Dynamic Color theme that is near grayscale.
"""

from PyMCUlib.dynamiccolor.color_spec_delegate import SpecVersion
from PyMCUlib.dynamiccolor.dynamic_scheme import DynamicScheme, Platform
from PyMCUlib.dynamiccolor.variant import Variant
from PyMCUlib.hct.hct import Hct


class SchemeNeutral(DynamicScheme):
    """A Dynamic Color theme that is near grayscale."""
    
    DEFAULT_SPEC_VERSION = '2021'
    DEFAULT_PLATFORM = 'phone'

    def __init__(
            self, 
            source_color_hct: Hct, 
            is_dark: bool, 
            contrast_level: float,
            spec_version: SpecVersion = DEFAULT_SPEC_VERSION,
            platform: Platform = DEFAULT_PLATFORM
    ):
        """
        Initialize a near grayscale theme with a source color.
        
        Args:
            source_color_hct: Source color in HCT color space.
            is_dark: Whether the theme is in dark mode.
            contrast_level: The contrast level of the theme.
            spec_version: The Material specification version to use. Defaults to DEFAULT_SPEC_VERSION.
            platform: The platform to use. Defaults to DEFAULT_PLATFORM.
        """            
        super().__init__({
            'source_color_hct': source_color_hct,
            'variant': Variant.NEUTRAL,
            'contrast_level': contrast_level,
            'is_dark': is_dark,
            'platform': platform,
            'spec_version': spec_version,
        })


# palettes/core_palette.py

"""
An intermediate concept between the key color for a UI theme, and a full
color scheme. 5 sets of tones are generated, all except one use the same hue
as the key color, and all vary in chroma.
"""

from typing import TypedDict, Optional

from PyMCUlib.hct.hct import Hct
from PyMCUlib.palettes.tonal_palette import TonalPalette


# Primary color is required; other colors are optional.
class CorePaletteColorsBase(TypedDict):
    """
    Colors needed to create a core palette.
    
    @deprecated Use DynamicScheme for color scheme generation.
    Use CorePalettes for core palettes container class.
    """
    primary: int

class CorePaletteColors(CorePaletteColorsBase, total=False):
    secondary: Optional[int]
    tertiary: Optional[int]
    neutral: Optional[int]
    neutralVariant: Optional[int]
    error: Optional[int]


class CorePalette:
    """
    An intermediate concept between the key color for a UI theme, and a full
    color scheme. 5 sets of tones are generated, all except one use the same hue
    as the key color, and all vary in chroma.
    
    @deprecated Use DynamicScheme for color scheme generation.
    Use CorePalettes for core palettes container class.
    """

    def __init__(self, argb: int, is_content: bool) -> None:
        """
        Initialize a CorePalette with a color and content flag.
        
        Args:
            argb: ARGB representation of a color
            is_content: Whether this palette is for content
        """
        hct = Hct.from_int(argb)
        hue = hct.hue
        chroma = hct.chroma
        
        if is_content:
            self.a1 = TonalPalette.from_hue_and_chroma(hue, chroma)
            self.a2 = TonalPalette.from_hue_and_chroma(hue, chroma / 3)
            self.a3 = TonalPalette.from_hue_and_chroma(hue + 60, chroma / 2)
            self.n1 = TonalPalette.from_hue_and_chroma(hue, min(chroma / 12, 4))
            self.n2 = TonalPalette.from_hue_and_chroma(hue, min(chroma / 6, 8))
        else:
            self.a1 = TonalPalette.from_hue_and_chroma(hue, max(48, chroma))
            self.a2 = TonalPalette.from_hue_and_chroma(hue, 16)
            self.a3 = TonalPalette.from_hue_and_chroma(hue + 60, 24)
            self.n1 = TonalPalette.from_hue_and_chroma(hue, 4)
            self.n2 = TonalPalette.from_hue_and_chroma(hue, 8)
            
        self.error = TonalPalette.from_hue_and_chroma(25, 84)

    @staticmethod
    def of(argb: int) -> 'CorePalette':
        """
        Create a standard CorePalette from a color.
        
        Args:
            argb: ARGB representation of a color
            
        Returns:
            Standard CorePalette
            
        @deprecated Use DynamicScheme for color scheme generation.
        Use CorePalettes for core palettes container class.
        """
        return CorePalette(argb, False)

    @staticmethod
    def content_of(argb: int) -> 'CorePalette':
        """
        Create a content CorePalette from a color.
        
        Args:
            argb: ARGB representation of a color
            
        Returns:
            Content CorePalette
            
        @deprecated Use DynamicScheme for color scheme generation.
        Use CorePalettes for core palettes container class.
        """
        return CorePalette(argb, True)

    @staticmethod
    def from_colors(colors: CorePaletteColors) -> 'CorePalette':
        """
        Create a CorePalette from a set of colors.
        
        Args:
            colors: CorePaletteColors
            
        Returns:
            Standard CorePalette
            
        @deprecated Use DynamicScheme for color scheme generation.
        Use CorePalettes for core palettes container class.
        """
        return CorePalette._create_palette_from_colors(False, colors)

    @staticmethod
    def content_from_colors(colors: CorePaletteColors) -> 'CorePalette':
        """
        Create a content CorePalette from a set of colors.
        
        Args:
            colors: CorePaletteColors
            
        Returns:
            Content CorePalette
            
        @deprecated Use DynamicScheme for color scheme generation.
        Use CorePalettes for core palettes container class.
        """
        return CorePalette._create_palette_from_colors(True, colors)

    @staticmethod
    def _create_palette_from_colors(
            content: bool,
            colors: CorePaletteColors) -> 'CorePalette':
        """
        Internal method to create a CorePalette from colors.
        
        Args:
            content: Whether this palette is for content
            colors: CorePaletteColors
            
        Returns:
            CorePalette
        """
        palette = CorePalette(colors["primary"], content)
        
        if "secondary" in colors and colors["secondary"] is not None:
            p = CorePalette(colors["secondary"], content)
            palette.a2 = p.a1
            
        if "tertiary" in colors and colors["tertiary"] is not None:
            p = CorePalette(colors["tertiary"], content)
            palette.a3 = p.a1
            
        if "error" in colors and colors["error"] is not None:
            p = CorePalette(colors["error"], content)
            palette.error = p.a1
            
        if "neutral" in colors and colors["neutral"] is not None:
            p = CorePalette(colors["neutral"], content)
            palette.n1 = p.n1
            
        if "neutralVariant" in colors and colors["neutralVariant"] is not None:
            p = CorePalette(colors["neutralVariant"], content)
            palette.n2 = p.n2
            
        return palette


# quantize/quantizer_wu.py

from typing import List
from PyMCUlib.utils import color_utils
from PyMCUlib.quantize.quantizer_map import QuantizerMap

# Constants
INDEX_BITS = 5
SIDE_LENGTH = 33  # ((1 << INDEX_INDEX_BITS) + 1)
TOTAL_SIZE = 35937  # SIDE_LENGTH * SIDE_LENGTH * SIDE_LENGTH

# Direction enum equivalent
class Direction:
    RED = "red"
    GREEN = "green"
    BLUE = "blue"


class Box:
    """
    Keeps track of the state of each box created as the Wu quantization
    algorithm progresses through dividing the image's pixels as plotted in RGB.
    """
    def __init__(
        self,
        r0: int = 0,
        r1: int = 0,
        g0: int = 0,
        g1: int = 0,
        b0: int = 0,
        b1: int = 0,
        vol: int = 0
    ) -> None:
        self.r0 = r0
        self.r1 = r1
        self.g0 = g0
        self.g1 = g1
        self.b0 = b0
        self.b1 = b1
        self.vol = vol


class CreateBoxesResult:
    """
    Represents final result of Wu algorithm.
    """
    def __init__(self, requested_count: int, result_count: int) -> None:
        """
        Args:
            requested_count: how many colors the caller asked to be returned from
                quantization.
            result_count: the actual number of colors achieved from quantization.
                May be lower than the requested count.
        """
        self.requested_count = requested_count
        self.result_count = result_count


class MaximizeResult:
    """
    Represents the result of calculating where to cut an existing box in such
    a way to maximize variance between the two new boxes created by a cut.
    """
    def __init__(self, cut_location: int, maximum: float) -> None:
        self.cut_location = cut_location
        self.maximum = maximum


class QuantizerWu:
    """
    An image quantizer that divides the image's pixels into clusters by
    recursively cutting an RGB cube, based on the weight of pixels in each area
    of the cube.

    The algorithm was described by Xiaolin Wu in Graphic Gems II, published in
    1991.
    """
    
    def __init__(
        self,
        weights: List[int] = None,
        moments_r: List[int] = None,
        moments_g: List[int] = None,
        moments_b: List[int] = None,
        moments: List[float] = None,
        cubes: List[Box] = None
    ) -> None:
        self.weights = weights if weights is not None else []
        self.moments_r = moments_r if moments_r is not None else []
        self.moments_g = moments_g if moments_g is not None else []
        self.moments_b = moments_b if moments_b is not None else []
        self.moments = moments if moments is not None else []
        self.cubes = cubes if cubes is not None else []

    def quantize(self, pixels: List[int], max_colors: int) -> List[int]:
        """
        Args:
            pixels: Colors in ARGB format.
            max_colors: The number of colors to divide the image into. A lower
                number of colors may be returned.
        
        Returns:
            Colors in ARGB format.
        """
        self._construct_histogram(pixels)
        self._compute_moments()
        create_boxes_result = self._create_boxes(max_colors)
        results = self._create_result(create_boxes_result.result_count)
        return results

    def _construct_histogram(self, pixels: List[int]) -> None:
        """
        Constructs a histogram of the image's pixels.
        """
        self.weights = [0] * TOTAL_SIZE
        self.moments_r = [0] * TOTAL_SIZE
        self.moments_g = [0] * TOTAL_SIZE
        self.moments_b = [0] * TOTAL_SIZE
        self.moments = [0] * TOTAL_SIZE

        count_by_color = QuantizerMap.quantize(pixels)

        for pixel, count in count_by_color.items():
            red = color_utils.red_from_argb(pixel)
            green = color_utils.green_from_argb(pixel)
            blue = color_utils.blue_from_argb(pixel)

            bits_to_remove = 8 - INDEX_BITS
            i_r = (red >> bits_to_remove) + 1
            i_g = (green >> bits_to_remove) + 1
            i_b = (blue >> bits_to_remove) + 1
            index = self._get_index(i_r, i_g, i_b)

            self.weights[index] = self.weights[index] + count
            self.moments_r[index] += count * red
            self.moments_g[index] += count * green
            self.moments_b[index] += count * blue
            self.moments[index] += count * (red * red + green * green + blue * blue)

    def _compute_moments(self) -> None:
        """
        Computes moments for the histogram.
        """
        for r in range(1, SIDE_LENGTH):
            area = [0] * SIDE_LENGTH
            area_r = [0] * SIDE_LENGTH
            area_g = [0] * SIDE_LENGTH
            area_b = [0] * SIDE_LENGTH
            area2 = [0.0] * SIDE_LENGTH
            
            for g in range(1, SIDE_LENGTH):
                line = 0
                line_r = 0
                line_g = 0
                line_b = 0
                line2 = 0.0
                
                for b in range(1, SIDE_LENGTH):
                    index = self._get_index(r, g, b)
                    line += self.weights[index]
                    line_r += self.moments_r[index]
                    line_g += self.moments_g[index]
                    line_b += self.moments_b[index]
                    line2 += self.moments[index]

                    area[b] += line
                    area_r[b] += line_r
                    area_g[b] += line_g
                    area_b[b] += line_b
                    area2[b] += line2

                    previous_index = self._get_index(r - 1, g, b)
                    self.weights[index] = self.weights[previous_index] + area[b]
                    self.moments_r[index] = self.moments_r[previous_index] + area_r[b]
                    self.moments_g[index] = self.moments_g[previous_index] + area_g[b]
                    self.moments_b[index] = self.moments_b[previous_index] + area_b[b]
                    self.moments[index] = self.moments[previous_index] + area2[b]

    def _create_boxes(self, max_colors: int) -> CreateBoxesResult:
        """
        Creates boxes for the quantization process.
        """
        self.cubes = [Box() for _ in range(max_colors)]
        volume_variance = [0.0] * max_colors
        
        self.cubes[0].r0 = 0
        self.cubes[0].g0 = 0
        self.cubes[0].b0 = 0

        self.cubes[0].r1 = SIDE_LENGTH - 1
        self.cubes[0].g1 = SIDE_LENGTH - 1
        self.cubes[0].b1 = SIDE_LENGTH - 1

        generated_color_count = max_colors
        next_index = 0
        
        i = 1
        while i < max_colors:
            # Try to cut cube[next_index], result in cube[i]
            if self._cut(self.cubes[next_index], self.cubes[i]):
                # Success, calculate the variance of the two new cubes
                volume_variance[next_index] = (
                    self._variance(self.cubes[next_index])
                    if self.cubes[next_index].vol > 1 else 0.0
                )
                volume_variance[i] = (
                    self._variance(self.cubes[i])
                    if self.cubes[i].vol > 1 else 0.0
                )
                increment_i = True
            else:
                # Failed, reset the variance of the cube, and do not move the index to try again
                volume_variance[next_index] = 0.0
                increment_i = False

            # Select the cube with the largest variance from all the cubes created (0 to i) as the next cutting target
            next_index = 0
            temp = volume_variance[0]
            for j in range(1, i + 1):
                if volume_variance[j] > temp:
                    temp = volume_variance[j]
                    next_index = j

            # If no further cutting is possible, record the actual number of colors generated and exit
            if temp <= 0.0:
                generated_color_count = i + 1
                break

            # Only move to the next i when the cutting is successful
            if increment_i:
                i += 1
                
        return CreateBoxesResult(max_colors, generated_color_count)

    def _create_result(self, color_count: int) -> List[int]:
        """
        Creates the final result colors from the boxes.
        """
        colors = []
        for i in range(color_count):
            cube = self.cubes[i]
            weight = self._volume(cube, self.weights)
            if weight > 0:
                r = round(self._volume(cube, self.moments_r) / weight)
                g = round(self._volume(cube, self.moments_g) / weight)
                b = round(self._volume(cube, self.moments_b) / weight)
                color = (255 << 24) | ((r & 0x0ff) << 16) | ((g & 0x0ff) << 8) | (b & 0x0ff)
                colors.append(color)
        return colors

    def _variance(self, cube: Box) -> float:
        """
        Calculates the variance for a box.
        """
        dr = self._volume(cube, self.moments_r)
        dg = self._volume(cube, self.moments_g)
        db = self._volume(cube, self.moments_b)
        xx = (
            self.moments[self._get_index(cube.r1, cube.g1, cube.b1)]
            - self.moments[self._get_index(cube.r1, cube.g1, cube.b0)]
            - self.moments[self._get_index(cube.r1, cube.g0, cube.b1)]
            + self.moments[self._get_index(cube.r1, cube.g0, cube.b0)]
            - self.moments[self._get_index(cube.r0, cube.g1, cube.b1)]
            + self.moments[self._get_index(cube.r0, cube.g1, cube.b0)]
            + self.moments[self._get_index(cube.r0, cube.g0, cube.b1)]
            - self.moments[self._get_index(cube.r0, cube.g0, cube.b0)]
        )
        hypotenuse = dr * dr + dg * dg + db * db
        volume = self._volume(cube, self.weights)
        return xx - hypotenuse / volume

    def _cut(self, one: Box, two: Box) -> bool:
        """
        Cuts a box in two.
        """
        whole_r = self._volume(one, self.moments_r)
        whole_g = self._volume(one, self.moments_g)
        whole_b = self._volume(one, self.moments_b)
        whole_w = self._volume(one, self.weights)

        max_r_result = self._maximize(
            one, Direction.RED, one.r0 + 1, one.r1, whole_r, whole_g, whole_b, whole_w
        )
        max_g_result = self._maximize(
            one, Direction.GREEN, one.g0 + 1, one.g1, whole_r, whole_g, whole_b, whole_w
        )
        max_b_result = self._maximize(
            one, Direction.BLUE, one.b0 + 1, one.b1, whole_r, whole_g, whole_b, whole_w
        )

        max_r = max_r_result.maximum
        max_g = max_g_result.maximum
        max_b = max_b_result.maximum

        if max_r >= max_g and max_r >= max_b:
            if max_r_result.cut_location < 0:
                return False
            direction = Direction.RED
        elif max_g >= max_r and max_g >= max_b:
            direction = Direction.GREEN
        else:
            direction = Direction.BLUE

        two.r1 = one.r1
        two.g1 = one.g1
        two.b1 = one.b1

        if direction == Direction.RED:
            one.r1 = max_r_result.cut_location
            two.r0 = one.r1
            two.g0 = one.g0
            two.b0 = one.b0
        elif direction == Direction.GREEN:
            one.g1 = max_g_result.cut_location
            two.r0 = one.r0
            two.g0 = one.g1
            two.b0 = one.b0
        elif direction == Direction.BLUE:
            one.b1 = max_b_result.cut_location
            two.r0 = one.r0
            two.g0 = one.g0
            two.b0 = one.b1
        else:
            raise ValueError(f"unexpected direction {direction}")

        one.vol = (one.r1 - one.r0) * (one.g1 - one.g0) * (one.b1 - one.b0)
        two.vol = (two.r1 - two.r0) * (two.g1 - two.g0) * (two.b1 - two.b0)
        return True

    def _maximize(
        self,
        cube: Box,
        direction: str,
        first: int,
        last: int,
        whole_r: int,
        whole_g: int,
        whole_b: int,
        whole_w: int,
    ) -> MaximizeResult:
        """
        Finds where to cut a box to maximize variance.
        """
        bottom_r = self._bottom(cube, direction, self.moments_r)
        bottom_g = self._bottom(cube, direction, self.moments_g)
        bottom_b = self._bottom(cube, direction, self.moments_b)
        bottom_w = self._bottom(cube, direction, self.weights)

        max_val = 0.0
        cut = -1

        half_r = 0
        half_g = 0
        half_b = 0
        half_w = 0
        
        for i in range(first, last):
            half_r = bottom_r + self._top(cube, direction, i, self.moments_r)
            half_g = bottom_g + self._top(cube, direction, i, self.moments_g)
            half_b = bottom_b + self._top(cube, direction, i, self.moments_b)
            half_w = bottom_w + self._top(cube, direction, i, self.weights)
            
            if half_w == 0:
                continue

            temp_numerator = (half_r * half_r + half_g * half_g + half_b * half_b) * 1.0
            temp_denominator = half_w * 1.0
            temp = temp_numerator / temp_denominator

            half_r = whole_r - half_r
            half_g = whole_g - half_g
            half_b = whole_b - half_b
            half_w = whole_w - half_w
            
            if half_w == 0:
                continue

            temp_numerator = (half_r * half_r + half_g * half_g + half_b * half_b) * 1.0
            temp_denominator = half_w * 1.0
            temp += temp_numerator / temp_denominator

            if temp > max_val:
                max_val = temp
                cut = i
                
        return MaximizeResult(cut, max_val)

    def _volume(self, cube: Box, moment: List[float]) -> float:
        """
        Calculates the volume of a box.
        """
        return (
            moment[self._get_index(cube.r1, cube.g1, cube.b1)]
            - moment[self._get_index(cube.r1, cube.g1, cube.b0)]
            - moment[self._get_index(cube.r1, cube.g0, cube.b1)]
            + moment[self._get_index(cube.r1, cube.g0, cube.b0)]
            - moment[self._get_index(cube.r0, cube.g1, cube.b1)]
            + moment[self._get_index(cube.r0, cube.g1, cube.b0)]
            + moment[self._get_index(cube.r0, cube.g0, cube.b1)]
            - moment[self._get_index(cube.r0, cube.g0, cube.b0)]
        )

    def _bottom(self, cube: Box, direction: str, moment: List[float]) -> float:
        """
        Calculates the bottom of a box.
        """
        if direction == Direction.RED:
            return (
                -moment[self._get_index(cube.r0, cube.g1, cube.b1)]
                + moment[self._get_index(cube.r0, cube.g1, cube.b0)]
                + moment[self._get_index(cube.r0, cube.g0, cube.b1)]
                - moment[self._get_index(cube.r0, cube.g0, cube.b0)]
            )
        elif direction == Direction.GREEN:
            return (
                -moment[self._get_index(cube.r1, cube.g0, cube.b1)]
                + moment[self._get_index(cube.r1, cube.g0, cube.b0)]
                + moment[self._get_index(cube.r0, cube.g0, cube.b1)]
                - moment[self._get_index(cube.r0, cube.g0, cube.b0)]
            )
        elif direction == Direction.BLUE:
            return (
                -moment[self._get_index(cube.r1, cube.g1, cube.b0)]
                + moment[self._get_index(cube.r1, cube.g0, cube.b0)]
                + moment[self._get_index(cube.r0, cube.g1, cube.b0)]
                - moment[self._get_index(cube.r0, cube.g0, cube.b0)]
            )
        else:
            raise ValueError(f"unexpected direction {direction}")

    def _top(self, cube: Box, direction: str, position: int, moment: List[float]) -> float:
        """
        Calculates the top of a box.
        """
        if direction == Direction.RED:
            return (
                moment[self._get_index(position, cube.g1, cube.b1)]
                - moment[self._get_index(position, cube.g1, cube.b0)]
                - moment[self._get_index(position, cube.g0, cube.b1)]
                + moment[self._get_index(position, cube.g0, cube.b0)]
            )
        elif direction == Direction.GREEN:
            return (
                moment[self._get_index(cube.r1, position, cube.b1)]
                - moment[self._get_index(cube.r1, position, cube.b0)]
                - moment[self._get_index(cube.r0, position, cube.b1)]
                + moment[self._get_index(cube.r0, position, cube.b0)]
            )
        elif direction == Direction.BLUE:
            return (
                moment[self._get_index(cube.r1, cube.g1, position)]
                - moment[self._get_index(cube.r1, cube.g0, position)]
                - moment[self._get_index(cube.r0, cube.g1, position)]
                + moment[self._get_index(cube.r0, cube.g0, position)]
            )
        else:
            raise ValueError(f"unexpected direction {direction}")

    def _get_index(self, r: int, g: int, b: int) -> int:
        """
        Gets the index for a given r,g,b coordinate.
        """
        return (r << (INDEX_BITS * 2)) + (r << (INDEX_BITS + 1)) + r + (g << INDEX_BITS) + g + b


# dynamiccolor/dynamic_scheme.py

from typing import List, Literal, Optional, TypedDict

from PyMCUlib.dynamiccolor.color_spec import get_spec
from PyMCUlib.dynamiccolor.color_spec_delegate import SpecVersion
from PyMCUlib.dynamiccolor.dynamic_color import DynamicColor
from PyMCUlib.dynamiccolor.material_dynamic_colors import MaterialDynamicColors
from PyMCUlib.dynamiccolor.variant import Variant
from PyMCUlib.dislike.dislike_analyzer import DislikeAnalyzer
from PyMCUlib.hct.hct import Hct
from PyMCUlib.palettes.tonal_palette import TonalPalette
from PyMCUlib.temperature.temperature_cache import TemperatureCache
from PyMCUlib.utils import math_utils


# The platform on which this scheme is intended to be used. Only used in the
# 2025 spec.
Platform = Literal['phone', 'watch']


class DynamicSchemeOptions(TypedDict, total=False):
    """
    Options for creating a DynamicScheme.
    
    Args:
        source_color_hct: The source color of the theme as an HCT color.
        variant: The variant, or style, of the theme.
        contrast_level: Value from -1 to 1. -1 represents minimum contrast, 0
            represents standard (i.e. the design as spec'd), and 1 represents maximum
            contrast.
        is_dark: Whether the scheme is in dark mode or light mode.
        platform: The platform on which this scheme is intended to be used.
        spec_version: The version of the design spec that this scheme is based on.
        primary_palette: Given a tone, produces a color. Hue and chroma of the
            color are specified in the design specification of the variant.
        secondary_palette: Given a tone, produces a color.
        tertiary_palette: Given a tone, produces a color.
        neutral_palette: Given a tone, produces a color.
        neutral_variant_palette: Given a tone, produces a color.
        error_palette: Given a tone, produces a color.
    """
    source_color_hct: Hct
    variant: Variant
    contrast_level: float
    is_dark: bool
    platform: Platform
    spec_version: SpecVersion
    primary_palette: TonalPalette
    secondary_palette: TonalPalette
    tertiary_palette: TonalPalette
    neutral_palette: TonalPalette
    neutral_variant_palette: TonalPalette
    error_palette: TonalPalette


class DynamicSchemePalettesDelegate:
    """
    A delegate that provides the palettes of a DynamicScheme.

    This is used to allow different implementations of the palette calculation
    logic for different spec versions.
    """
    def get_primary_palette(
            self, variant: Variant, source_color_hct: Hct, is_dark: bool,
            platform: Platform, contrast_level: float) -> TonalPalette:
        """
        Get the primary palette for the given parameters.
        
        Args:
            variant: The variant.
            source_color_hct: The source color.
            is_dark: Whether the scheme is in dark mode.
            platform: The platform.
            contrast_level: The contrast level.
            
        Returns:
            A tonal palette.
        """
        ...

    def get_secondary_palette(
            self, variant: Variant, source_color_hct: Hct, is_dark: bool,
            platform: Platform, contrast_level: float) -> TonalPalette:
        """
        Get the secondary palette for the given parameters.
        
        Args:
            variant: The variant.
            source_color_hct: The source color.
            is_dark: Whether the scheme is in dark mode.
            platform: The platform.
            contrast_level: The contrast level.
            
        Returns:
            A tonal palette.
        """
        ...

    def get_tertiary_palette(
            self, variant: Variant, source_color_hct: Hct, is_dark: bool,
            platform: Platform, contrast_level: float) -> TonalPalette:
        """
        Get the tertiary palette for the given parameters.
        
        Args:
            variant: The variant.
            source_color_hct: The source color.
            is_dark: Whether the scheme is in dark mode.
            platform: The platform.
            contrast_level: The contrast level.
            
        Returns:
            A tonal palette.
        """
        ...

    def get_neutral_palette(
            self, variant: Variant, source_color_hct: Hct, is_dark: bool,
            platform: Platform, contrast_level: float) -> TonalPalette:
        """
        Get the neutral palette for the given parameters.
        
        Args:
            variant: The variant.
            source_color_hct: The source color.
            is_dark: Whether the scheme is in dark mode.
            platform: The platform.
            contrast_level: The contrast level.
            
        Returns:
            A tonal palette.
        """
        ...

    def get_neutral_variant_palette(
            self, variant: Variant, source_color_hct: Hct, is_dark: bool,
            platform: Platform, contrast_level: float) -> TonalPalette:
        """
        Get the neutral variant palette for the given parameters.
        
        Args:
            variant: The variant.
            source_color_hct: The source color.
            is_dark: Whether the scheme is in dark mode.
            platform: The platform.
            contrast_level: The contrast level.
            
        Returns:
            A tonal palette.
        """
        ...

    def get_error_palette(
            self, variant: Variant, source_color_hct: Hct, is_dark: bool,
            platform: Platform, contrast_level: float) -> Optional[TonalPalette]:
        """
        Get the error palette for the given parameters.
        
        Args:
            variant: The variant.
            source_color_hct: The source color.
            is_dark: Whether the scheme is in dark mode.
            platform: The platform.
            contrast_level: The contrast level.
            
        Returns:
            A tonal palette, or None if there's no error palette.
        """
        ...


class DynamicScheme:
    """
    Constructed by a set of values representing the current UI state (such as
    whether or not its dark theme, what the theme style is, etc.), and
    provides a set of TonalPalettes that can create colors that fit in
    with the theme style. Used by DynamicColor to resolve into a color.
    """

    def __init__(self, args: DynamicSchemeOptions):
        """
        Initialize a dynamic scheme.
        
        Args:
            args: Options for creating a DynamicScheme.
        """
        self.source_color_argb = args['source_color_hct'].to_int()
        self.variant = args['variant']
        self.contrast_level = args['contrast_level']
        self.is_dark = args['is_dark']
        self.platform = args.get('platform', 'phone')
        self.spec_version = args.get('spec_version', '2021')
        self.source_color_hct = args['source_color_hct']
        self.primary_palette = args.get('primary_palette') or get_spec(self.spec_version).get_primary_palette(
            self.variant, args['source_color_hct'], self.is_dark, self.platform,
            self.contrast_level)
        self.secondary_palette = args.get('secondary_palette') or get_spec(self.spec_version).get_secondary_palette(
            self.variant, args['source_color_hct'], self.is_dark, self.platform,
            self.contrast_level)
        self.tertiary_palette = args.get('tertiary_palette') or get_spec(self.spec_version).get_tertiary_palette(
            self.variant, args['source_color_hct'], self.is_dark, self.platform,
            self.contrast_level)
        self.neutral_palette = args.get('neutral_palette') or get_spec(self.spec_version).get_neutral_palette(
            self.variant, args['source_color_hct'], self.is_dark, self.platform,
            self.contrast_level)
        self.neutral_variant_palette = args.get('neutral_variant_palette') or get_spec(
            self.spec_version).get_neutral_variant_palette(
            self.variant, args['source_color_hct'], self.is_dark, self.platform,
            self.contrast_level)
        error_palette = args.get('error_palette') or get_spec(self.spec_version).get_error_palette(
            self.variant, args['source_color_hct'], self.is_dark, self.platform,
            self.contrast_level)
        self.error_palette = error_palette if error_palette else TonalPalette.from_hue_and_chroma(25.0, 84.0)
        
        self.colors = MaterialDynamicColors()

    def __str__(self) -> str:
        """
        Returns a string representation of the scheme.
        
        Returns:
            A string representation of the scheme.
        """
        return (
            f"Scheme: variant={Variant(self.variant).name}, "
            f"mode={'dark' if self.is_dark else 'light'}, "
            f"platform={self.platform}, "
            f"contrastLevel={self.contrast_level:.1f}, "
            f"seed={self.source_color_hct}, "
            f"specVersion={self.spec_version}"
        )

    @staticmethod
    def get_piecewise_hue(
            source_color_hct: Hct, hue_breakpoints: List[float], hues: List[float]) -> float:
        """
        Returns a new hue based on a piecewise function and input color hue.

        For example, for the following function:
        result = 26 if 0 <= hue < 101
        result = 39 if 101 <= hue < 210
        result = 28 if 210 <= hue < 360

        call the function as:

        hue_breakpoints = [0, 101, 210, 360]
        hues = [26, 39, 28]
        result = scheme.piecewise(hue, hue_breakpoints, hues)

        Args:
            source_color_hct: The input value.
            hue_breakpoints: The breakpoints, in sorted order. No default lower or
                upper bounds are assumed.
            hues: The hues that should be applied when source color's hue is >=
                the same index in hueBrakpoints array, and < the hue at the next index
                in hueBrakpoints array. Otherwise, the source color's hue is returned.
                
        Returns:
            A hue value.
        """
        size = min(len(hue_breakpoints) - 1, len(hues))
        source_hue = source_color_hct.hue
        for i in range(size):
            if source_hue >= hue_breakpoints[i] and source_hue < hue_breakpoints[i + 1]:
                return math_utils.sanitize_degrees_double(hues[i])
        # No condition matched, return the source hue.
        return source_hue

    @staticmethod
    def get_rotated_hue(
            source_color_hct: Hct, hue_breakpoints: List[float],
            rotations: List[float]) -> float:
        """
        Returns a shifted hue based on a piecewise function and input color hue.

        For example, for the following function:
        result = hue + 26 if 0 <= hue < 101
        result = hue - 39 if 101 <= hue < 210
        result = hue + 28 if 210 <= hue < 360

        call the function as:

        hue_breakpoints = [0, 101, 210, 360]
        hues = [26, -39, 28]
        result = scheme.get_rotated_hue(hue, hue_breakpoints, hues)

        Args:
            source_color_hct: the source color of the theme, in HCT.
            hue_breakpoints: The "breakpoints", i.e. the hues at which a rotation
                should be apply. No default lower or upper bounds are assumed.
            rotations: The rotation that should be applied when source color's
                hue is >= the same index in hues array, and < the hue at the next
                index in hues array. Otherwise, the source color's hue is returned.
                
        Returns:
            A hue value.
        """
        rotation = DynamicScheme.get_piecewise_hue(
            source_color_hct, hue_breakpoints, rotations)
        if min(len(hue_breakpoints) - 1, len(rotations)) <= 0:
            # No condition matched, return the source hue.
            rotation = 0
        return math_utils.sanitize_degrees_double(source_color_hct.hue + rotation)

    def get_argb(self, dynamic_color: DynamicColor) -> int:
        """
        Get the ARGB color for the given dynamic color.
        
        Args:
            dynamic_color: The dynamic color.
            
        Returns:
            An ARGB color as an integer.
        """
        return dynamic_color.get_argb(self)

    def get_hct(self, dynamic_color: DynamicColor) -> Hct:
        """
        Get the HCT color for the given dynamic color.
        
        Args:
            dynamic_color: The dynamic color.
            
        Returns:
            An HCT color.
        """
        return dynamic_color.get_hct(self)

    # Palette key colors

    @property
    def primary_palette_key_color(self) -> int:
        """
        Get the primary palette key color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.primary_palette_key_color())

    @property
    def secondary_palette_key_color(self) -> int:
        """
        Get the secondary palette key color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.secondary_palette_key_color())

    @property
    def tertiary_palette_key_color(self) -> int:
        """
        Get the tertiary palette key color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.tertiary_palette_key_color())

    @property
    def neutral_palette_key_color(self) -> int:
        """
        Get the neutral palette key color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.neutral_palette_key_color())

    @property
    def neutral_variant_palette_key_color(self) -> int:
        """
        Get the neutral variant palette key color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.neutral_variant_palette_key_color())

    @property
    def error_palette_key_color(self) -> int:
        """
        Get the error palette key color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.error_palette_key_color())

    # Surface colors

    @property
    def background(self) -> int:
        """
        Get the background color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.background())

    @property
    def on_background(self) -> int:
        """
        Get the on-background color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.on_background())

    @property
    def surface(self) -> int:
        """
        Get the surface color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.surface())

    @property
    def surface_dim(self) -> int:
        """
        Get the dim surface color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.surface_dim())

    @property
    def surface_bright(self) -> int:
        """
        Get the bright surface color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.surface_bright())

    @property
    def surface_container_lowest(self) -> int:
        """
        Get the lowest surface container color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.surface_container_lowest())

    @property
    def surface_container_low(self) -> int:
        """
        Get the low surface container color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.surface_container_low())

    @property
    def surface_container(self) -> int:
        """
        Get the surface container color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.surface_container())

    @property
    def surface_container_high(self) -> int:
        """
        Get the high surface container color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.surface_container_high())

    @property
    def surface_container_highest(self) -> int:
        """
        Get the highest surface container color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.surface_container_highest())

    @property
    def on_surface(self) -> int:
        """
        Get the on-surface color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.on_surface())

    @property
    def surface_variant(self) -> int:
        """
        Get the surface variant color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.surface_variant())

    @property
    def on_surface_variant(self) -> int:
        """
        Get the on-surface variant color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.on_surface_variant())

    @property
    def inverse_surface(self) -> int:
        """
        Get the inverse surface color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.inverse_surface())

    @property
    def inverse_on_surface(self) -> int:
        """
        Get the inverse on-surface color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.inverse_on_surface())

    @property
    def outline(self) -> int:
        """
        Get the outline color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.outline())

    @property
    def outline_variant(self) -> int:
        """
        Get the outline variant color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.outline_variant())

    @property
    def shadow(self) -> int:
        """
        Get the shadow color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.shadow())

    @property
    def scrim(self) -> int:
        """
        Get the scrim color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.scrim())

    @property
    def surface_tint(self) -> int:
        """
        Get the surface tint color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.surface_tint())

    # Primary colors

    @property
    def primary(self) -> int:
        """
        Get the primary color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.primary())

    @property
    def primary_dim(self) -> int:
        """
        Get the dim primary color.
        
        Returns:
            An ARGB color as an integer.
            
        Raises:
            ValueError: If the primary_dim color is undefined.
        """
        primary_dim = self.colors.primary_dim()
        if primary_dim is None:
            raise ValueError("`primary_dim` color is undefined prior to 2025 spec.")
        return self.get_argb(primary_dim)

    @property
    def on_primary(self) -> int:
        """
        Get the on-primary color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.on_primary())

    @property
    def primary_container(self) -> int:
        """
        Get the primary container color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.primary_container())

    @property
    def on_primary_container(self) -> int:
        """
        Get the on-primary container color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.on_primary_container())

    @property
    def primary_fixed(self) -> int:
        """
        Get the fixed primary color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.primary_fixed)

    @property
    def primary_fixed_dim(self) -> int:
        """
        Get the dim fixed primary color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.primary_fixed_dim)

    @property
    def on_primary_fixed(self) -> int:
        """
        Get the on-fixed primary color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.on_primary_fixed)

    @property
    def on_primary_fixed_variant(self) -> int:
        """
        Get the on-fixed variant primary color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.on_primary_fixed_variant)

    @property
    def inverse_primary(self) -> int:
        """
        Get the inverse primary color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.inverse_primary())

    # Secondary colors

    @property
    def secondary(self) -> int:
        """
        Get the secondary color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.secondary())

    @property
    def secondary_dim(self) -> int:
        """
        Get the dim secondary color.
        
        Returns:
            An ARGB color as an integer.
            
        Raises:
            ValueError: If the secondary_dim color is undefined.
        """
        secondary_dim = self.colors.secondary_dim()
        if secondary_dim is None:
            raise ValueError("`secondary_dim` color is undefined prior to 2025 spec.")
        return self.get_argb(secondary_dim)

    @property
    def on_secondary(self) -> int:
        """
        Get the on-secondary color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.on_secondary())

    @property
    def secondary_container(self) -> int:
        """
        Get the secondary container color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.secondary_container())

    @property
    def on_secondary_container(self) -> int:
        """
        Get the on-secondary container color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.on_secondary_container())

    @property
    def secondary_fixed(self) -> int:
        """
        Get the fixed secondary color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.secondary_fixed)

    @property
    def secondary_fixed_dim(self) -> int:
        """
        Get the dim fixed secondary color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.secondary_fixed_dim)

    @property
    def on_secondary_fixed(self) -> int:
        """
        Get the on-fixed secondary color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.on_secondary_fixed)

    @property
    def on_secondary_fixed_variant(self) -> int:
        """
        Get the on-fixed variant secondary color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.on_secondary_fixed_variant)

    # Tertiary colors

    @property
    def tertiary(self) -> int:
        """
        Get the tertiary color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.tertiary())

    @property
    def tertiary_dim(self) -> int:
        """
        Get the dim tertiary color.
        
        Returns:
            An ARGB color as an integer.
            
        Raises:
            ValueError: If the tertiary_dim color is undefined.
        """
        tertiary_dim = self.colors.tertiary_dim()
        if tertiary_dim is None:
            raise ValueError("`tertiary_dim` color is undefined prior to 2025 spec.")
        return self.get_argb(tertiary_dim)

    @property
    def on_tertiary(self) -> int:
        """
        Get the on-tertiary color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.on_tertiary())

    @property
    def tertiary_container(self) -> int:
        """
        Get the tertiary container color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.tertiary_container())

    @property
    def on_tertiary_container(self) -> int:
        """
        Get the on-tertiary container color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.on_tertiary_container())

    @property
    def tertiary_fixed(self) -> int:
        """
        Get the fixed tertiary color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.tertiary_fixed)

    @property
    def tertiary_fixed_dim(self) -> int:
        """
        Get the dim fixed tertiary color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.tertiary_fixed_dim)

    @property
    def on_tertiary_fixed(self) -> int:
        """
        Get the on-fixed tertiary color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.on_tertiary_fixed)

    @property
    def on_tertiary_fixed_variant(self) -> int:
        """
        Get the on-fixed variant tertiary color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.on_tertiary_fixed_variant)

    # Error colors

    @property
    def error(self) -> int:
        """
        Get the error color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.error())

    @property
    def error_dim(self) -> int:
        """
        Get the dim error color.
        
        Returns:
            An ARGB color as an integer.
            
        Raises:
            ValueError: If the error_dim color is undefined.
        """
        error_dim = self.colors.error_dim()
        if error_dim is None:
            raise ValueError("`error_dim` color is undefined prior to 2025 spec.")
        return self.get_argb(error_dim)

    @property
    def on_error(self) -> int:
        """
        Get the on-error color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.on_error())

    @property
    def error_container(self) -> int:
        """
        Get the error container color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.error_container())

    @property
    def on_error_container(self) -> int:
        """
        Get the on-error container color.
        
        Returns:
            An ARGB color as an integer.
        """
        return self.get_argb(self.colors.on_error_container())


class DynamicSchemePalettesDelegateImpl2021(DynamicSchemePalettesDelegate):
    """
    A delegate for the palettes of a DynamicScheme in the 2021 spec.
    """

    def get_primary_palette(
            self, variant: Variant, source_color_hct: Hct, is_dark: bool,
            platform: Platform, contrast_level: float) -> TonalPalette:
        """
        Get the primary palette for the given parameters.
        
        Args:
            variant: The variant.
            source_color_hct: The source color.
            is_dark: Whether the scheme is in dark mode.
            platform: The platform.
            contrast_level: The contrast level.
            
        Returns:
            A tonal palette.
        """
        if variant == Variant.CONTENT or variant == Variant.FIDELITY:
            return TonalPalette.from_hue_and_chroma(
                source_color_hct.hue, source_color_hct.chroma)
        elif variant == Variant.FRUIT_SALAD:
            return TonalPalette.from_hue_and_chroma(
                math_utils.sanitize_degrees_double(source_color_hct.hue - 50.0), 48.0)
        elif variant == Variant.MONOCHROME:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 0.0)
        elif variant == Variant.NEUTRAL:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 12.0)
        elif variant == Variant.RAINBOW:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 48.0)
        elif variant == Variant.TONAL_SPOT:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 36.0)
        elif variant == Variant.EXPRESSIVE:
            return TonalPalette.from_hue_and_chroma(
                math_utils.sanitize_degrees_double(source_color_hct.hue + 240), 40)
        elif variant == Variant.VIBRANT:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 200.0)
        else:
            raise ValueError(f"Unsupported variant: {variant}")

    def get_secondary_palette(
            self, variant: Variant, source_color_hct: Hct, is_dark: bool,
            platform: Platform, contrast_level: float) -> TonalPalette:
        """
        Get the secondary palette for the given parameters.
        
        Args:
            variant: The variant.
            source_color_hct: The source color.
            is_dark: Whether the scheme is in dark mode.
            platform: The platform.
            contrast_level: The contrast level.
            
        Returns:
            A tonal palette.
        """
        if variant == Variant.CONTENT or variant == Variant.FIDELITY:
            return TonalPalette.from_hue_and_chroma(
                source_color_hct.hue,
                max(source_color_hct.chroma - 32.0, source_color_hct.chroma * 0.5))
        elif variant == Variant.FRUIT_SALAD:
            return TonalPalette.from_hue_and_chroma(
                math_utils.sanitize_degrees_double(source_color_hct.hue - 50.0), 36.0)
        elif variant == Variant.MONOCHROME:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 0.0)
        elif variant == Variant.NEUTRAL:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 8.0)
        elif variant == Variant.RAINBOW:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 16.0)
        elif variant == Variant.TONAL_SPOT:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 16.0)
        elif variant == Variant.EXPRESSIVE:
            return TonalPalette.from_hue_and_chroma(
                DynamicScheme.get_rotated_hue(
                    source_color_hct, [0, 21, 51, 121, 151, 191, 271, 321, 360],
                    [45, 95, 45, 20, 45, 90, 45, 45, 45]),
                24.0)
        elif variant == Variant.VIBRANT:
            return TonalPalette.from_hue_and_chroma(
                DynamicScheme.get_rotated_hue(
                    source_color_hct, [0, 41, 61, 101, 131, 181, 251, 301, 360],
                    [18, 15, 10, 12, 15, 18, 15, 12, 12]),
                24.0)
        else:
            raise ValueError(f"Unsupported variant: {variant}")

    def get_tertiary_palette(
            self, variant: Variant, source_color_hct: Hct, is_dark: bool,
            platform: Platform, contrast_level: float) -> TonalPalette:
        """
        Get the tertiary palette for the given parameters.
        
        Args:
            variant: The variant.
            source_color_hct: The source color.
            is_dark: Whether the scheme is in dark mode.
            platform: The platform.
            contrast_level: The contrast level.
            
        Returns:
            A tonal palette.
        """
        if variant == Variant.CONTENT:
            return TonalPalette.from_hct(DislikeAnalyzer.fix_if_disliked(
                TemperatureCache(source_color_hct).analogous(3, 6)[2]))
        elif variant == Variant.FIDELITY:
            return TonalPalette.from_hct(DislikeAnalyzer.fix_if_disliked(
                TemperatureCache(source_color_hct).complement))
        elif variant == Variant.FRUIT_SALAD:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 36.0)
        elif variant == Variant.MONOCHROME:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 0.0)
        elif variant == Variant.NEUTRAL:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 16.0)
        elif variant == Variant.RAINBOW or variant == Variant.TONAL_SPOT:
            return TonalPalette.from_hue_and_chroma(
                math_utils.sanitize_degrees_double(source_color_hct.hue + 60.0), 24.0)
        elif variant == Variant.EXPRESSIVE:
            return TonalPalette.from_hue_and_chroma(
                DynamicScheme.get_rotated_hue(
                    source_color_hct, [0, 21, 51, 121, 151, 191, 271, 321, 360],
                    [120, 120, 20, 45, 20, 15, 20, 120, 120]),
                32.0)
        elif variant == Variant.VIBRANT:
            return TonalPalette.from_hue_and_chroma(
                DynamicScheme.get_rotated_hue(
                    source_color_hct, [0, 41, 61, 101, 131, 181, 251, 301, 360],
                    [35, 30, 20, 25, 30, 35, 30, 25, 25]),
                32.0)
        else:
            raise ValueError(f"Unsupported variant: {variant}")

    def get_neutral_palette(
            self, variant: Variant, source_color_hct: Hct, is_dark: bool,
            platform: Platform, contrast_level: float) -> TonalPalette:
        """
        Get the neutral palette for the given parameters.
        
        Args:
            variant: The variant.
            source_color_hct: The source color.
            is_dark: Whether the scheme is in dark mode.
            platform: The platform.
            contrast_level: The contrast level.
            
        Returns:
            A tonal palette.
        """
        if variant == Variant.CONTENT or variant == Variant.FIDELITY:
            return TonalPalette.from_hue_and_chroma(
                source_color_hct.hue, source_color_hct.chroma / 8.0)
        elif variant == Variant.FRUIT_SALAD:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 10.0)
        elif variant == Variant.MONOCHROME:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 0.0)
        elif variant == Variant.NEUTRAL:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 2.0)
        elif variant == Variant.RAINBOW:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 0.0)
        elif variant == Variant.TONAL_SPOT:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 6.0)
        elif variant == Variant.EXPRESSIVE:
            return TonalPalette.from_hue_and_chroma(
                math_utils.sanitize_degrees_double(source_color_hct.hue + 15), 8)
        elif variant == Variant.VIBRANT:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 10)
        else:
            raise ValueError(f"Unsupported variant: {variant}")

    def get_neutral_variant_palette(
            self, variant: Variant, source_color_hct: Hct, is_dark: bool,
            platform: Platform, contrast_level: float) -> TonalPalette:
        """
        Get the neutral variant palette for the given parameters.
        
        Args:
            variant: The variant.
            source_color_hct: The source color.
            is_dark: Whether the scheme is in dark mode.
            platform: The platform.
            contrast_level: The contrast level.
            
        Returns:
            A tonal palette.
        """
        if variant == Variant.CONTENT:
            return TonalPalette.from_hue_and_chroma(
                source_color_hct.hue, (source_color_hct.chroma / 8.0) + 4.0)
        elif variant == Variant.FIDELITY:
            return TonalPalette.from_hue_and_chroma(
                source_color_hct.hue, (source_color_hct.chroma / 8.0) + 4.0)
        elif variant == Variant.FRUIT_SALAD:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 16.0)
        elif variant == Variant.MONOCHROME:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 0.0)
        elif variant == Variant.NEUTRAL:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 2.0)
        elif variant == Variant.RAINBOW:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 0.0)
        elif variant == Variant.TONAL_SPOT:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 8.0)
        elif variant == Variant.EXPRESSIVE:
            return TonalPalette.from_hue_and_chroma(
                math_utils.sanitize_degrees_double(source_color_hct.hue + 15), 12)
        elif variant == Variant.VIBRANT:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 12)
        else:
            raise ValueError(f"Unsupported variant: {variant}")

    def get_error_palette(
            self, variant: Variant, source_color_hct: Hct, is_dark: bool,
            platform: Platform, contrast_level: float) -> Optional[TonalPalette]:
        """
        Get the error palette for the given parameters.
        
        Args:
            variant: The variant.
            source_color_hct: The source color.
            is_dark: Whether the scheme is in dark mode.
            platform: The platform.
            contrast_level: The contrast level.
            
        Returns:
            None, as there's no error palette in the 2021 spec.
        """
        return None


class DynamicSchemePalettesDelegateImpl2025(DynamicSchemePalettesDelegateImpl2021):
    """
    A delegate for the palettes of a DynamicScheme in the 2025 spec.
    """

    def get_primary_palette(
            self, variant: Variant, source_color_hct: Hct, is_dark: bool,
            platform: Platform, contrast_level: float) -> TonalPalette:
        """
        Get the primary palette for the given parameters.
        
        Args:
            variant: The variant.
            source_color_hct: The source color.
            is_dark: Whether the scheme is in dark mode.
            platform: The platform.
            contrast_level: The contrast level.
            
        Returns:
            A tonal palette.
        """
        if variant == Variant.NEUTRAL:
            return TonalPalette.from_hue_and_chroma(
                source_color_hct.hue,
                12 if platform == 'phone' and Hct.is_blue(source_color_hct.hue) else
                8 if platform == 'phone' else
                16 if Hct.is_blue(source_color_hct.hue) else 12)
        elif variant == Variant.TONAL_SPOT:
            return TonalPalette.from_hue_and_chroma(
                source_color_hct.hue, 26 if platform == 'phone' and is_dark else 32)
        elif variant == Variant.EXPRESSIVE:
            return TonalPalette.from_hue_and_chroma(
                source_color_hct.hue, 36 if platform == 'phone' and is_dark else
                48 if platform == 'phone' else 40)
        elif variant == Variant.VIBRANT:
            return TonalPalette.from_hue_and_chroma(
                source_color_hct.hue, 74 if platform == 'phone' else 56)
        else:
            return super().get_primary_palette(
                variant, source_color_hct, is_dark, platform, contrast_level)

    def get_secondary_palette(
            self, variant: Variant, source_color_hct: Hct, is_dark: bool,
            platform: Platform, contrast_level: float) -> TonalPalette:
        """
        Get the secondary palette for the given parameters.
        
        Args:
            variant: The variant.
            source_color_hct: The source color.
            is_dark: Whether the scheme is in dark mode.
            platform: The platform.
            contrast_level: The contrast level.
            
        Returns:
            A tonal palette.
        """
        if variant == Variant.NEUTRAL:
            return TonalPalette.from_hue_and_chroma(
                source_color_hct.hue,
                6 if platform == 'phone' and Hct.is_blue(source_color_hct.hue) else
                4 if platform == 'phone' else
                10 if Hct.is_blue(source_color_hct.hue) else 6)
        elif variant == Variant.TONAL_SPOT:
            return TonalPalette.from_hue_and_chroma(source_color_hct.hue, 16)
        elif variant == Variant.EXPRESSIVE:
            return TonalPalette.from_hue_and_chroma(
                DynamicScheme.get_rotated_hue(
                    source_color_hct, [0, 105, 140, 204, 253, 278, 300, 333, 360],
                    [-160, 155, -100, 96, -96, -156, -165, -160]),
                16 if platform == 'phone' and is_dark else 24 if platform == 'phone' else 24)
        elif variant == Variant.VIBRANT:
            return TonalPalette.from_hue_and_chroma(
                DynamicScheme.get_rotated_hue(
                    source_color_hct, [0, 38, 105, 140, 333, 360],
                    [-14, 10, -14, 10, -14]),
                56 if platform == 'phone' else 36)
        else:
            return super().get_secondary_palette(
                variant, source_color_hct, is_dark, platform, contrast_level)

    def get_tertiary_palette(
            self, variant: Variant, source_color_hct: Hct, is_dark: bool,
            platform: Platform, contrast_level: float) -> TonalPalette:
        """
        Get the tertiary palette for the given parameters.
        
        Args:
            variant: The variant.
            source_color_hct: The source color.
            is_dark: Whether the scheme is in dark mode.
            platform: The platform.
            contrast_level: The contrast level.
            
        Returns:
            A tonal palette.
        """
        if variant == Variant.NEUTRAL:
            return TonalPalette.from_hue_and_chroma(
                DynamicScheme.get_rotated_hue(
                    source_color_hct, [0, 38, 105, 161, 204, 278, 333, 360],
                    [-32, 26, 10, -39, 24, -15, -32]),
                20 if platform == 'phone' else 36)
        elif variant == Variant.TONAL_SPOT:
            return TonalPalette.from_hue_and_chroma(
                DynamicScheme.get_rotated_hue(
                    source_color_hct, [0, 20, 71, 161, 333, 360],
                    [-40, 48, -32, 40, -32]),
                28 if platform == 'phone' else 32)
        elif variant == Variant.EXPRESSIVE:
            return TonalPalette.from_hue_and_chroma(
                DynamicScheme.get_rotated_hue(
                    source_color_hct, [0, 105, 140, 204, 253, 278, 300, 333, 360],
                    [-165, 160, -105, 101, -101, -160, -170, -165]),
                48)
        elif variant == Variant.VIBRANT:
            return TonalPalette.from_hue_and_chroma(
                DynamicScheme.get_rotated_hue(
                    source_color_hct, [0, 38, 71, 105, 140, 161, 253, 333, 360],
                    [-72, 35, 24, -24, 62, 50, 62, -72]),
                56)
        else:
            return super().get_tertiary_palette(
                variant, source_color_hct, is_dark, platform, contrast_level)

    @staticmethod
    def _get_expressive_neutral_hue(source_color_hct: Hct) -> float:
        """
        Get the expressive neutral hue for the given source color.
        
        Args:
            source_color_hct: The source color.
            
        Returns:
            A hue value.
        """
        hue = DynamicScheme.get_rotated_hue(
            source_color_hct, [0, 71, 124, 253, 278, 300, 360],
            [10, 0, 10, 0, 10, 0])
        return hue

    @staticmethod
    def _get_expressive_neutral_chroma(
            source_color_hct: Hct, is_dark: bool, platform: Platform) -> float:
        """
        Get the expressive neutral chroma for the given parameters.
        
        Args:
            source_color_hct: The source color.
            is_dark: Whether the scheme is in dark mode.
            platform: The platform.
            
        Returns:
            A chroma value.
        """
        neutral_hue = DynamicSchemePalettesDelegateImpl2025._get_expressive_neutral_hue(
            source_color_hct)
        return (6 if platform == 'phone' and is_dark and Hct.is_yellow(neutral_hue) else
               14 if platform == 'phone' and is_dark else
               18 if platform == 'phone' else 12)

    @staticmethod
    def _get_vibrant_neutral_hue(source_color_hct: Hct) -> float:
        """
        Get the vibrant neutral hue for the given source color.
        
        Args:
            source_color_hct: The source color.
            
        Returns:
            A hue value.
        """
        return DynamicScheme.get_rotated_hue(
            source_color_hct, [0, 38, 105, 140, 333, 360], [-14, 10, -14, 10, -14])

    @staticmethod
    def _get_vibrant_neutral_chroma(source_color_hct: Hct, platform: Platform) -> float:
        """
        Get the vibrant neutral chroma for the given parameters.
        
        Args:
            source_color_hct: The source color.
            platform: The platform.
            
        Returns:
            A chroma value.
        """
        neutral_hue = DynamicSchemePalettesDelegateImpl2025._get_vibrant_neutral_hue(
            source_color_hct)
        return (28 if platform == 'phone' else
               28 if Hct.is_blue(neutral_hue) else 20)

    def get_neutral_palette(
            self, variant: Variant, source_color_hct: Hct, is_dark: bool,
            platform: Platform, contrast_level: float) -> TonalPalette:
        """
        Get the neutral palette for the given parameters.
        
        Args:
            variant: The variant.
            source_color_hct: The source color.
            is_dark: Whether the scheme is in dark mode.
            platform: The platform.
            contrast_level: The contrast level.
            
        Returns:
            A tonal palette.
        """
        if variant == Variant.NEUTRAL:
            return TonalPalette.from_hue_and_chroma(
                source_color_hct.hue, 1.4 if platform == 'phone' else 6)
        elif variant == Variant.TONAL_SPOT:
            return TonalPalette.from_hue_and_chroma(
                source_color_hct.hue, 5 if platform == 'phone' else 10)
        elif variant == Variant.EXPRESSIVE:
            return TonalPalette.from_hue_and_chroma(
                self._get_expressive_neutral_hue(source_color_hct),
                self._get_expressive_neutral_chroma(source_color_hct, is_dark, platform))
        elif variant == Variant.VIBRANT:
            return TonalPalette.from_hue_and_chroma(
                self._get_vibrant_neutral_hue(source_color_hct),
                self._get_vibrant_neutral_chroma(source_color_hct, platform))
        else:
            return super().get_neutral_palette(
                variant, source_color_hct, is_dark, platform, contrast_level)

    def get_neutral_variant_palette(
            self, variant: Variant, source_color_hct: Hct, is_dark: bool,
            platform: Platform, contrast_level: float) -> TonalPalette:
        """
        Get the neutral variant palette for the given parameters.
        
        Args:
            variant: The variant.
            source_color_hct: The source color.
            is_dark: Whether the scheme is in dark mode.
            platform: The platform.
            contrast_level: The contrast level.
            
        Returns:
            A tonal palette.
        """
        if variant == Variant.NEUTRAL:
            return TonalPalette.from_hue_and_chroma(
                source_color_hct.hue, (1.4 if platform == 'phone' else 6) * 2.2)
        elif variant == Variant.TONAL_SPOT:
            return TonalPalette.from_hue_and_chroma(
                source_color_hct.hue, (5 if platform == 'phone' else 10) * 1.7)
        elif variant == Variant.EXPRESSIVE:
            expressive_neutral_hue = self._get_expressive_neutral_hue(source_color_hct)
            expressive_neutral_chroma = self._get_expressive_neutral_chroma(
                source_color_hct, is_dark, platform)
            return TonalPalette.from_hue_and_chroma(
                expressive_neutral_hue,
                expressive_neutral_chroma * (
                    1.6 if expressive_neutral_hue >= 105 and expressive_neutral_hue < 125 else 2.3),
            )
        elif variant == Variant.VIBRANT:
            vibrant_neutral_hue = self._get_vibrant_neutral_hue(source_color_hct)
            vibrant_neutral_chroma = self._get_vibrant_neutral_chroma(source_color_hct, platform)
            return TonalPalette.from_hue_and_chroma(
                vibrant_neutral_hue, vibrant_neutral_chroma * 1.29)
        else:
            return super().get_neutral_variant_palette(
                variant, source_color_hct, is_dark, platform, contrast_level)

    def get_error_palette(
            self, variant: Variant, source_color_hct: Hct, is_dark: bool,
            platform: Platform, contrast_level: float) -> Optional[TonalPalette]:
        """
        Get the error palette for the given parameters.
        
        Args:
            variant: The variant.
            source_color_hct: The source color.
            is_dark: Whether the scheme is in dark mode.
            platform: The platform.
            contrast_level: The contrast level.
            
        Returns:
            A tonal palette.
        """
        error_hue = DynamicScheme.get_piecewise_hue(
            source_color_hct, [0, 3, 13, 23, 33, 43, 153, 273, 360],
            [12, 22, 32, 12, 22, 32, 22, 12])
        if variant == Variant.NEUTRAL:
            return TonalPalette.from_hue_and_chroma(
                error_hue, 50 if platform == 'phone' else 40)
        elif variant == Variant.TONAL_SPOT:
            return TonalPalette.from_hue_and_chroma(
                error_hue, 60 if platform == 'phone' else 48)
        elif variant == Variant.EXPRESSIVE:
            return TonalPalette.from_hue_and_chroma(
                error_hue, 64 if platform == 'phone' else 48)
        elif variant == Variant.VIBRANT:
            return TonalPalette.from_hue_and_chroma(
                error_hue, 80 if platform == 'phone' else 60)
        else:
            return super().get_error_palette(
                variant, source_color_hct, is_dark, platform, contrast_level)


# Global delegates for different spec versions
spec2021 = DynamicSchemePalettesDelegateImpl2021()
spec2025 = DynamicSchemePalettesDelegateImpl2025()


def get_spec(spec_version: SpecVersion) -> DynamicSchemePalettesDelegate:
    """
    Returns the DynamicSchemePalettesDelegate for the given spec version.
    
    Args:
        spec_version: The spec version.
        
    Returns:
        A DynamicSchemePalettesDelegate.
    """
    return spec2025 if spec_version == '2025' else spec2021